/*
 *     Refactoring.java  Mar 25, 2004
 *
 *     Katsuhisa Maruyama (maru@jtool.org)
 */

package org.jtool.jrbx.refactor;
import org.jtool.jrbx.refactor.util.*;
import org.jtool.jrbx.refactor.dialog.*;
import org.jtool.jrbx.gui.Refactor;
import org.jtool.platform.xml.dom.*;
import org.jtool.platform.xml.wrapper.*;
import org.jtool.platform.j2xml.*;
import java.util.*;
import javax.swing.*;
import org.w3c.dom.*;

public abstract class Refactoring implements RefactoringCommand {
    protected static String commandName;
    protected Document document;
    protected Element elem; 
    protected ArrayList elemIds;
    protected ArrayList selectedElems;
    protected Refactor refactor;
    protected JFrame frame = null;
    protected JavaFile jfile;
    protected ArrayList changedCodes = new ArrayList();  // DisplayedCode
    private ArrayList changedXMLTexts = new ArrayList();  // JavaXMLText
    private String logMessage = null;

    protected Refactoring() {
    }

    public static RefactoringCommand create(String command, String packagePath) {
        commandName = command;

        try {
            Class cl = Class.forName(packagePath + "." + command);
            return (RefactoringCommand)cl.newInstance();

        } catch (ClassNotFoundException e) {
            // e.printStackTrace();
        } catch (InstantiationException e) {
            // e.printStackTrace();
        } catch (IllegalAccessException e) {
            // e.printStackTrace();
        }
        return new NullCommand();
    }

    public void setSource(Document doc, ArrayList ids, ArrayList elems) {
        document = doc;
        elemIds = ids;
        selectedElems = elems;

        elem = GetElem.getElem(doc, (String)elemIds.get(0));
        Repository.setDescendants();
    }

    public void setRefactor(Refactor refactor) {
        this.refactor = refactor;
        if (refactor != null) {
            frame = refactor.getFrame();
        }
    }

    public ArrayList getChangedCodes() {
        return changedCodes;
    }

    public ArrayList getChangedXMLTexts() {
        return changedXMLTexts;
    }

    public String getRefactoringLog() {
        return logMessage;
    }

    public void execute() {
        try {
            setUp();
            preconditions();
            transform();
            additionalTransformation();

            if (confirm()) {
                acceptCode();
                logMessage = commandName + ": " + getLog();

            } else {
                recoverCode();
                changedCodes.clear();
                logMessage = null;
            }

        } catch (RefactoringException e) {
            recover(e);
        }
    }

    protected abstract void setUp();
    
    protected abstract void preconditions() throws RefactoringException;

    protected abstract void transform() throws RefactoringException;

    protected abstract void additionalTransformation() throws RefactoringException;
    
    protected abstract String getLog();

    protected boolean confirm() {
        boolean result = PreviewDialog.show(frame, "Preview Changes", changedCodes);
        return result;
    }

    protected void recover(RefactoringException e) {
        String mesg = e.getMessage();
        if (mesg.length() != 3) {
            JOptionPane.showMessageDialog(frame, commandName + mesg);
        }

        recoverCode();
        changedCodes.clear();
        logMessage = null;
    }

    protected void unparse(JavaFile jfile) {
        removeCode(jfile.getName());
        // System.out.println(jfile);

        Document doc = jfile.getElem().getOwnerDocument();
        String oldCode = Dom2Code.getOldCode(doc);
        ArrayList oldHighlights = Dom2Code.getHighlights();

        String newCode = Dom2Code.getNewCode(doc);
        ArrayList newHighlights = Dom2Code.getHighlights();

        DisplayedCode code = new DisplayedCode(jfile.getName(), oldCode, newCode);
        code.setOldHighlight(oldHighlights);
        code.setNewHighlight(newHighlights);
        changedCodes.add(code);
    }

    private void removeCode(String fileName) {
        ArrayList tmpCodes = new ArrayList(changedCodes);
        Iterator it = tmpCodes.iterator();
        while (it.hasNext()) {
            DisplayedCode code = (DisplayedCode)it.next();
            if (code.getName().equals(fileName)) {
                changedCodes.remove(code);
            }
        }
    }

    private void acceptCode() {
        Iterator it = changedCodes.iterator();
        while (it.hasNext()) {
            DisplayedCode code = (DisplayedCode)it.next();

            JavaXMLText text = Repository.getJavaXMLText(code.getName());
            text.setSourceCodeOnEdit(code.getNewCode());
            changedXMLTexts.add(text);
        }
    }

    private void recoverCode() {
        Iterator it = changedCodes.iterator();
        while (it.hasNext()) {
            DisplayedCode code = (DisplayedCode)it.next();
            JavaXMLText text = Repository.getJavaXMLText(code.getName());
            text.refresh();
        }
    }

    protected String getRelativeName(String name) {
        String projectDir = SystemProperties.getProjectDir();
        if (name != null && name.startsWith(projectDir)) {
            return name.substring(projectDir.length() + 1);
        }
        return "UNKNOWN";
    }

    protected boolean isInProjectDir(JavaFile jfile) {
        return SystemProperties.isInProjectDir(jfile.getName());
    }

    protected void print(JavaFile jfile) {
        Document doc = jfile.getElem().getOwnerDocument();
        String t = org.jtool.platform.xml.dom.DOMTreeWriter.getXMLText(doc);
        System.out.println(t);
    }
}
