/*
 *     ExtractMethod.java  Feb 23, 2004
 *
 *     Katsuhisa Maruyama (maru@jtool.org)
 */

package org.jtool.jrbx.refactor.miscellaneous;
import org.jtool.jrbx.refactor.*;
import org.jtool.jrbx.refactor.util.*;
import org.jtool.jrbx.refactor.query.*;
import org.jtool.jrbx.refactor.dialog.*;
import org.jtool.platform.xml.JXConstants;
import org.jtool.platform.xml.dom.*;
import org.jtool.platform.xml.wrapper.*;
import org.jtool.platform.xml.graph.*;
import org.jtool.platform.xml.graph.cfg.*;
import org.jtool.platform.xml.graph.pdg.*;
import org.w3c.dom.*;
import java.util.*;
import javax.swing.JFrame;

public class ExtractMethod extends MiscellaneousRefactoring {
    private JavaStatement jstatement;
    private String newName = null;
    private JavaVariable rjv = null;
    private ExtractMethodTransformer transformer;
    private MethodCode mcode;
    private CFG cfg;
    private PDG pdg;
    private GraphCompSet innerPDGNodes;
    private GraphCompSet outerPDGNodes;

    public ExtractMethod() {
        super();
    }

    public void setUpTest(JFrame f, String newName) {
        frame = f;
        this.newName = newName;
    }

    protected void setUp() {
        elem = getFirstStatementElement();
        if (elem != null) {
            jstatement = new JavaStatement(elem);
            super.setUp();
        }
    }

    private Element getFirstStatementElement() {
        Iterator it = selectedElems.iterator();
        while (it.hasNext()) {
            Element e = (Element)it.next();

            while (e != null &&
                   !e.getNodeName().equals(JXConstants.StatementElem) &&
                   !e.getNodeName().equals(JXConstants.LocalElem) &&
                   !e.getNodeName().equals(JXConstants.LabelElem)) {
                e = (Element)e.getParentNode();
            }
            if (e != null) {
                return e;
            }
        }
        return null;
    }

    protected void preconditions() throws RefactoringException {
        if (elem == null) {
            throw new RefactoringException("Not extract:" +
              " the selection does not contains a statement");
        }

        if (!isEnclosedBySingleMethod(selectedElems)) {
            throw new RefactoringException("Not extract:" +
              " the selection is not enclosed by the single method");
        }

        cfg = new CFG(jstatement.getJavaMethod().getElem());
        pdg = new PDG(cfg);
        collectPDGNodes();

        if (containsThisOrSuperCall(innerPDGNodes)) {
            throw new RefactoringException("Not extract:" +
              " the selection contains this or super call");
        }

        if (containsBranchWithUnselectedTarget(innerPDGNodes)) {
            throw new RefactoringException("Not extract:" +
              " the selection contains a break or continue statement whose target is not selected");
        }

        if (!isEnclosedByTheSameParent(innerPDGNodes)) {
            throw new RefactoringException("Not extract:" +
              " the selection is not enclosed by the single parent statement");
        }

        JavaVariableList vars = getReturnJavaVariables();
        String returnType = "void";
        if (vars.size() > 1) {
            throw new RefactoringException("Not extract:" +
              " the selection contains more than one return variable");
        } else if (vars.size() == 1) {
            rjv = vars.getFirst();
            returnType = rjv.getType();
        }

        if (newName == null) {
            newName = NewNameDialog.show(frame, "Extract Method");
        }
        if (newName == null) {
            throw new RefactoringException("");
        }

        mcode = new MethodCode(returnType, newName);
        collectParameters(mcode);
        if (QueryClass.hasTheMethod(jclass, mcode.getSignature())) {
            throw new RefactoringException("Not rename:" +
              " the method " + newName + " already exists in the class " + jclass.getName());
        }
    }

    protected void transform() throws RefactoringException {
        transformer = new ExtractMethodTransformer(jstatement, mcode, cfg, rjv, innerPDGNodes);
        transformer.changeSrc();
        unparse(jfile);
    }

    protected void additionalTransformation() throws RefactoringException {
    }

    protected String getLog() {
        return mcode.getName() + " from " + jstatement.getJavaMethod().getName();
    }

    private void collectPDGNodes() {
        innerPDGNodes = getSelectedPDGNodes();
        outerPDGNodes = pdg.getNodes().difference(innerPDGNodes);
    }

    private GraphCompSet getSelectedPDGNodes() {
        GraphCompSet nodes = new GraphCompSet();

        Iterator it = selectedElems.iterator();
        while (it.hasNext()) {
            Element e = (Element)it.next();

            collectPDGNodes(e, nodes);
        }
        return nodes;
    }

    private void collectPDGNodes(Element e, GraphCompSet nodes) {
        while (e != null) {
            PDGNode pdgNode = getPDGNode(e);
            if (pdgNode != null) {
                nodes.add(pdgNode);
            }

            if (e.getNodeName().equals(JXConstants.StatementElem) ||
                e.getNodeName().equals(JXConstants.LocalElem) ||
                e.getNodeName().equals(JXConstants.LabelElem)) {
                return;

            } else {
                Node parent = e.getParentNode();
                if (parent != null && parent.getNodeType() == Node.ELEMENT_NODE) {
                    e = (Element)parent;
                } else {
                    return;
                }
            }
        }
    }

    private PDGNode getPDGNode(Element e) {
        CFGNode cfgNode = cfg.getNode(e);

        Iterator it = pdg.getNodes().iterator();
        while (it.hasNext()) {
            PDGNode node = (PDGNode)it.next();

            if (node.getCFGNode().equals(cfgNode)) {
                return node;
            }
        }
        return null;
    }

    private JavaVariableList getReturnJavaVariables() {
        JavaVariableList vars = new JavaVariableList();

        Iterator it = pdg.getEdges().iterator();
        while (it.hasNext()) {
            Dependence edge = (Dependence)it.next();

            if (edge.isDD()) {
                DD dd = (DD)edge;
                if (innerPDGNodes.contains(dd.getSrcNode()) &&
                    outerPDGNodes.contains(dd.getDstNode())) {

                    JavaVariable jv =  dd.getVariable();
                    if ((jv.isLocal() || jv.isParameter() || jv.isMethodReturn()) &&
                        !contains(vars, jv)) {
                        vars.add(jv);
                    }
                }
            }
        }
        return vars;
    }

    private void collectParameters(MethodCode code) {
        JavaVariableList vars = new JavaVariableList();

        Iterator it = pdg.getEdges().iterator();
        while (it.hasNext()) {
            Dependence edge = (Dependence)it.next();

            if (edge.isDD()) {
                DD dd = (DD)edge;
                if (outerPDGNodes.contains(dd.getSrcNode()) &&
                    innerPDGNodes.contains(dd.getDstNode())) {

                    JavaVariable jv =  dd.getVariable();
                    if ((jv.isLocal() || jv.isParameter()) && !contains(vars, jv)) {
                        vars.add(jv);
                        code.addParameter(jv.getType(), jv.getTypeFqn(), jv.getName());
                    }
                }
            }
        }
    }

    private boolean contains(JavaVariableList vars, JavaVariable jvar) {
        Iterator it = vars.iterator();
        while (it.hasNext()) {
            JavaVariable jv = (JavaVariable)it.next();

            if (jv.getName().equals(jvar.getName())) {
                return true;
            }
        }
        return false;
    }

    private ArrayList collectElements(GraphCompSet set) {
        ArrayList nodes = new ArrayList();

        Iterator it = set.iterator();
        while (it.hasNext()) {
            PDGNode node = (PDGNode)it.next();
            nodes.add(node.getCFGNode().getElem());
        }
        return nodes;
    }

    private boolean isEnclosedBySingleMethod(ArrayList elems) {
        Iterator it = selectedElems.iterator();
        while (it.hasNext()) {
            Element e = (Element)it.next();

            JavaElement jelem = new JavaElement(e);
            jmethod = jelem.getJavaMethod(e);
            if (jmethod == null || jelem.isMethodOrConstructorNode(e)) {
                return false;
            }
        }
        return true;
    }

    private boolean containsThisOrSuperCall(GraphCompSet nodes) {
        Iterator it = nodes.iterator();
        while (it.hasNext()) {
            PDGNode pdgNode = (PDGNode)it.next();

            CFGNode cfgNode = pdgNode.getCFGNode();
            if (cfgNode.isCallSt()) {
                CFGCall n = (CFGCall)cfgNode;
                if (n.getName().equals("this") || n.getName().equals("super")) {
                    return true;
                }
            }
        }
        return false;
    }

    private boolean containsBranchWithUnselectedTarget(GraphCompSet nodes) {
        Iterator it = nodes.iterator();
        while (it.hasNext()) {
            PDGNode pdgNode = (PDGNode)it.next();

            CFGNode cfgNode = pdgNode.getCFGNode();
            if (cfgNode.isBreakSt() || cfgNode.isContinueSt()) {
                CFGNode target = getTargetNode(cfgNode);

                if (target != null) {
                    PDGNode n = pdg.getNode(target);
                    if (n != null && !nodes.contains(n)) {
                        return true;
                    }
                }
            }
        }
        return false;
    }

    private CFGNode getTargetNode(CFGNode node) {
        Iterator it = node.getOutgoingEdges().iterator();
        while (it.hasNext()) {
            Flow flow = (Flow)it.next();
            if (flow.isTrue()) {
                return (CFGNode)flow.getDstNode();
            }
        }
        return null;
    }

    private boolean isEnclosedByTheSameParent(GraphCompSet nodes) {
        GraphCompSet parentNodes = new GraphCompSet();

        Iterator it = getTopNodes(nodes).iterator();
        while (it.hasNext()) {
            PDGNode pdgNode = (PDGNode)it.next();

            Iterator ite = pdgNode.getIncomingCDEdges().iterator();
            while (ite.hasNext()) {
                CD cd = (CD)ite.next();
                parentNodes.add(cd.getSrcNode());
            }
        }
        return parentNodes.size() == 1;
    }
    
    private GraphCompSet getTopNodes(GraphCompSet nodes) {
        GraphCompSet topNodes = new GraphCompSet(nodes);

        Iterator it = nodes.iterator();
        while (it.hasNext()) {
            PDGNode pdgNode = (PDGNode)it.next();
            
            Iterator ite = pdgNode.getIncomingCDEdges().iterator();
            while (ite.hasNext()) {
                CD cd = (CD)ite.next();

                if (nodes.contains(cd.getSrcNode())) {
                    topNodes.remove(pdgNode);
                }
            }
        }
        return topNodes;
    }

    private void print(GraphCompSet set) {
        Iterator it = set.iterator();
        while (it.hasNext()) {
            GraphNode node = (GraphNode)it.next();
            node.print();
        }
    }
}
