/*
 *     ExtractMethodTransformer.java  Feb 17, 2004
 *
 *     Katsuhisa Maruyama (maru@jtool.org)
 */

package org.jtool.jrbx.refactor.miscellaneous;
import org.jtool.jrbx.refactor.*;
import org.jtool.jrbx.refactor.util.*;
import org.jtool.platform.xml.wrapper.*;
import org.jtool.platform.xml.graph.*;
import org.jtool.platform.xml.graph.cfg.*;
import org.jtool.platform.xml.graph.pdg.*;
import java.util.*;
import org.w3c.dom.*;

public class ExtractMethodTransformer extends RefactoringTransformer {
    private JavaStatement jstatement;
    private MethodCode mcode;
    private CFG cfg;
    private JavaVariable rjv;
    private ArrayList innerElems = new ArrayList();
    private GraphCompSet innerCFGNodes = new GraphCompSet();

    public ExtractMethodTransformer(JavaStatement jstatement, MethodCode mcode,
                                    CFG cfg, JavaVariable rjv, GraphCompSet nodes) {
        this.jstatement = jstatement;
        this.mcode = mcode;
        this.cfg = cfg;
        this.rjv = rjv;

        collectElements(nodes);
    }

    private void collectElements(GraphCompSet nodes) {
        Iterator it = nodes.iterator();
        while (it.hasNext()) {
            PDGNode node = (PDGNode)it.next();

            innerCFGNodes.add(node.getCFGNode());
            innerElems.add(node.getCFGNode().getElem());
        }
    }

    public void changeSrc() {
        insertDeclaration();
        String code = createExtractedMethodCode();
        insertCodeAfter("\n\n" + code, jstatement.getJavaMethod().getElem());

        deleteExtractedMethodCode();
        insertMethodCall();
    }

    private void insertMethodCall() {
        StringBuffer buf = new StringBuffer();

        if (rjv != null) {
            if (rjv.isMethodReturn()) {
                buf.append("return ");
            } else {
                if (!isDeclared()) {
                    buf.append(rjv.getType());
                    buf.append(" ");
                }
                buf.append(rjv.getName());
                buf.append(" = ");
            }
        }

        buf.append(mcode.getName());
        buf.append("(");
        buf.append(mcode.getParameterNames());
        buf.append(");");

        insertCodeBefore(buf.toString(), jstatement.getElem());
    }

    private String createExtractedMethodCode() {
        if (jstatement.getJavaMethod().isStatic()) {
            mcode.addModifier(Modifier.STATIC);
        }
        mcode.addModifier(Modifier.PRIVATE);

        StringBuffer buf = new StringBuffer();
        buf.append("    ");
        buf.append(mcode.getCode());
        buf.append(" {\n");
        buf.append(getContents());
        
        if (rjv != null && needReturn()) {
            buf.append("        ");
            buf.append("return ");
            buf.append(rjv.getName());
            buf.append(";\n");
        }
        buf.append("    }");

        return buf.toString();
    }

    private String getContents() {
        StringBuffer buf = new StringBuffer();

        Iterator it = innerElems.iterator();
        while(it.hasNext()) {
            Element elem = (Element)it.next();

            if (!containsParents(innerElems, elem.getParentNode())) {
                buf.append(getContentsOfCode(elem));

                JavaElement jelem = new JavaElement(elem);
                if (jelem.isFieldNode(elem) || jelem.isLocalNode(elem)) {
                    buf.append(";");
                }
                buf.append("\n");
            }
        }
        return buf.toString();
    }

    private boolean containsParents(ArrayList nodes, Node node) {
        while (node.getNodeType() == Node.ELEMENT_NODE && node != null) {
            if (contains(nodes, (Element)node)) {
                return true;
            }
            node = node.getParentNode();
        }
        return false;
    }

    private boolean contains(ArrayList nodes, Element elem) {
        Iterator it = nodes.iterator();
        while (it.hasNext()) {
            Element e = (Element)it.next();

            if (e.equals(elem)) {
                return true;
            }
        }
        return false;
    }

    private boolean needReturn() {
        Iterator it = innerCFGNodes.iterator();
        while(it.hasNext()) {
            CFGNode cfgNode = (CFGNode)it.next();

            Iterator ite = cfgNode.getOutgoingEdges().iterator();
            while (ite.hasNext()) {
                Flow flow = (Flow)ite.next();

                if (flow.isTrue() || flow.isFalse()) {
                    CFGNode succ = (CFGNode)flow.getDstNode();

                    if (!succ.isFormalSt() && !innerCFGNodes.contains(succ)) {
                        return true;
                    }
                }
            }
        }
        return false;
    }

    private boolean isDeclared() {
        CFGNode node = cfg.getNode(jstatement.getElem());
        GraphCompSet path = cfg.getBackwardReachableNodesWithoutLoopback(node, cfg.getStartNode());
        path.remove(node);
        
        Iterator it = path.iterator();
        while(it.hasNext()) {
            CFGNode cfgNode = (CFGNode)it.next();

            if (cfgNode.isLocalSt()) {
                CFGDeclaration declNode = (CFGDeclaration)cfgNode;
                JavaVariable jv = (JavaVariable)declNode.getDeclaration();

                if (jv.equals(rjv)) {
                    return true;
                }
            }
        }
        return false;
    }

    private void deleteExtractedMethodCode() {
        Iterator it = innerElems.iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();

            JavaElement jelem = new JavaElement(elem);
            if (jelem.isFieldNode(elem)) {
                delete(new JavaField(elem));
            } else if (jelem.isLocalNode(elem)) {
                delete(new JavaLocal(elem));

            } else {
                delete(elem);
            }
        }
    }

    private CFGNode traverseBackward(GraphCompSet candidates) {
        GraphCompSet path = cfg.getNodes();

        Iterator it = candidates.iterator();
        while (it.hasNext()) {
            PDGNode n = (PDGNode)it.next();
            CFGNode node = n.getCFGNode();

            GraphCompSet p =
                cfg.getBackwardReachableNodesWithoutLoopback(node, cfg.getStartNode());
            path = path.intersection(p);
        }
        return (CFGNode)path.getFirst();
    }

    private void insertDeclaration() {
        Iterator it = innerCFGNodes.iterator();
        while(it.hasNext()) {
            CFGNode cfgNode = (CFGNode)it.next();

            if (cfgNode.hasDefVariable() && !cfgNode.isLocalSt()) {
                CFGStatement node = (CFGStatement)cfgNode;

                JavaVariable jv = node.getDefVars().getFirst();
                if (jv != null) {
                    Element e = (Element)jv.getDeclaration();

                    if (e != null) { 
                        CFGNode declNode = cfg.getNode(e);

                        if (declNode.isLocalSt()) {
                            if (!innerCFGNodes.contains(declNode)) {
                                insertCodeBefore(jv.getType() + " ", jv.getElem());
                                System.out.println("NEEDS DECl " + jv.getName());
                            }
                        }
                    }
                }
            }
        }
    }

    private void print(GraphCompSet set) {
        Iterator it = set.iterator();
        while (it.hasNext()) {
            GraphNode node = (GraphNode)it.next();
           node.print();
        }
    }
}
