/*
 *     SplitVariableTransformer.java  Feb 12, 2004
 *
 *     Katsuhisa Maruyama (maru@jtool.org)
 */

package org.jtool.jrbx.refactor.variables;
import org.jtool.jrbx.refactor.util.*;
import org.jtool.platform.xml.wrapper.*;
import org.jtool.platform.xml.graph.*;
import org.jtool.platform.xml.graph.cfg.*;
import org.jtool.platform.xml.graph.pdg.*;
import java.util.*;
import org.w3c.dom.*;

public class SplitVariableTransformer extends RefactoringTransformer {
    private JavaVariable jvar;
    private String newName;
    private CFG cfg;
    private PDG pdg;
    private GraphCompSet nodesToBeChanged;

    private SplitVariableTransformer() {
    }

    public SplitVariableTransformer(JavaVariable jvar, CFG cfg, PDG pdg,
                                    GraphCompSet cnodes, String newName) {
        this.jvar = jvar;
        this.newName = newName;
        this.cfg = cfg;
        this.pdg = pdg;
        nodesToBeChanged = cnodes;
    }

    public void changeSrc() {
        CFGNode decl = getDeclaraionCFGNode(nodesToBeChanged);
        changeVariableName(nodesToBeChanged, decl);
    }

    private CFGNode getDeclaraionCFGNode(GraphCompSet cnodes) {
        GraphCompSet candidates = getCandidatesForDeclaration(cnodes);

        CFGNode cfgNode;
        if (candidates.size() == 1) {
            PDGNode pdgNode = (PDGNode)candidates.getFirst();
            cfgNode = pdgNode.getCFGNode();
        } else {
            cfgNode = traverseBackward(candidates);
        }

        return cfgNode;
    }

    private GraphCompSet getCandidatesForDeclaration(GraphCompSet cnodes) {
        GraphCompSet candidates = new GraphCompSet();

        Iterator it = cnodes.iterator();
        while (it.hasNext()) {
            PDGNode node = (PDGNode)it.next();

            PDGStatement n = (PDGStatement)node;
            if (n.containsDefVariable(jvar)) {
                candidates.add(node);
            }
        }
        return candidates;
    }

    private CFGNode traverseBackward(GraphCompSet candidates) {
        GraphCompSet path = cfg.getNodes();
        
        Iterator it = candidates.iterator();
        while (it.hasNext()) {
            PDGNode n = (PDGNode)it.next();
            CFGNode node = n.getCFGNode();

            GraphCompSet p =
                cfg.getBackwardReachableNodesWithoutLoopback(node, cfg.getStartNode());
            path = path.intersection(p);
        }
        return (CFGNode)path.getFirst();
    }

    private void changeVariableName(GraphCompSet cnodes, CFGNode decl) {
        Iterator it = pdg.getNodes().iterator();
        while (it.hasNext()) {
            PDGNode pdgNode = (PDGNode)it.next();
            CFGNode cfgNode = pdgNode.getCFGNode();

            boolean isContained = cnodes.contains(pdgNode);
            boolean needsType = insertDeclaration(cfgNode, decl, isContained);

            if (isContained) {
                JavaStatement jst = new JavaStatement(cfgNode.getElem());
                if (needsType) {
                    String type = jvar.getType();
                    changeUsingName(jst.getDefVariables(), type + " " + newName);
                } else {
                    changeUsingName(jst.getDefVariables(), newName);
                }

                changeUsingName(jst.getUseVariables(), newName);
            }
        }
    }

    private boolean insertDeclaration(CFGNode node, CFGNode decl, boolean isContained) {
        if (node.equals(decl)) {
            if (isContained) {
            
                if (isDefinedAt(decl, jvar)) {
                    if (decl.getSort() != GraphNodeSort.localDecl) {
                        return true;
                    }
                    
                } else {
                    insertDeclaration(decl);
                }
                
            } else {
                insertDeclaration(decl);
            }
        }
        return false;
    }

    private boolean isDefinedAt(CFGNode node, JavaVariable jv) {
        if (node.hasDefVariable()) {
            CFGStatement n = (CFGStatement)node;
            if (n.containsDefVariable(jv)) {
                return true;
            }
        }
        return false;
    }

    private void insertDeclaration(CFGNode node) {
        String space = getSpace(node.getElem());
        String declStr = jvar.getType() + " " + newName + ";\n" + space;
        insertCodeBefore(declStr, node.getElem());
    }

    private void changeUsingName(JavaVariableList vars, String name) {
        Iterator it = vars.iterator();
        while (it.hasNext()) {
            JavaVariable jv = (JavaVariable)it.next();

            if ((jv.isLocal() || jv.isParameter()) &&
                jv.getName().equals(jvar.getName())) {
                rename(name, jv);
            }
        }
    }

    private void print(GraphCompSet set) {
        Iterator it = set.iterator();
        while (it.hasNext()) {
            GraphNode node = (GraphNode)it.next();
            node.print();
        }
    }
}
